/*****************************************************
* Filename: tms.c
* Description: Standard Turing Machine Simulator.
*
* (c) 1997 Jeff Stefan
*******************************************************/
#include <stdio.h>
#include <string.h>
#include <conio.h>

#define MAX_TRANSITIONS 30
#define MAX_TAPE_LEN 50
#define MAX_FILENAME_LEN 13
#define MAX_FILE_LINE_LEN 80

#define SINGLE_STEP 1
#define AUTO        2

/**********************
* Function prototypes
***********************/
int ReadTape();
int DisplayTape();
int DisplayTransitions(int,int);
int ReadTuringMachine();
int ExecuteTransitions();
void PressEnterMsg();

/**************************
* Turing Machine structure
***************************/
typedef struct tm
{
  int State;
  char Symbol;
  int NewState;
  char NewSymbol;
  char MoveTo;
} TURING_MACHINE;
TURING_MACHINE TM[MAX_TRANSITIONS];
int NumTransitions = 0;

/***************************
*  Single Step or Auto mode
****************************/
int Mode = 0;

/******************
* Input tape array
******************/
char InputTape[MAX_TAPE_LEN];

/***************
* File pointers
****************/
FILE *Tapefp;
FILE *TMfp;
char Filename[MAX_FILENAME_LEN];

void main(void)
{
int Selection = 0;
int ch;

  clrscr();
  printf("\n\nTMS: Turing Machine Simulator\n");
  printf("----------------------------------\n");
  printf(" 1: Single Step\n");
  printf(" 2: Auto\n");
  printf("----------------------------------\n");
  printf("Make Selection: ");
  Selection = getchar();
  Selection = abs(48 -Selection);
  switch(Selection)
  {
    case 1: /* Single Step Mode */
      Mode = SINGLE_STEP;
      ch = getchar();
      break;

    case 2: /* Auto Mode */
      Mode = AUTO;
      ch = getchar();
      break;

    default:
      Mode = AUTO;
      ch = getchar();
      break;
  }

  ReadTape();
  ReadTuringMachine();
  ExecuteTransitions();
  DisplayTape();
 }

/*****************************************************
*:: ReadTape
*
******************************************************/
int ReadTape()
{
int i;
char TapeFileName[MAX_FILENAME_LEN];
char ch;

  /****************************
  * init tape to all blanks B
  *****************************/
  for(i=0;i < MAX_TAPE_LEN;i++)
    InputTape[i] = 'B';
  i=0;

  /**********************
  * Open input tape file
  ***********************/
  printf("Enter Input Tape Filename: ");
  gets(TapeFileName);
  printf("\n");
  Tapefp = fopen(TapeFileName,"r");

  /****************************
  * exit if file doesn't exist
  ****************************/
  if(Tapefp == NULL)
  {
    printf("No Input Tape\n");
    exit(0);
  }

  /***************************************************
  * read tape symbols starting from first B to last B
  and load InputTape array.
  ****************************************************/
  while(!feof(Tapefp))
  {
    ch =  fgetc(Tapefp);
    switch(ch)
    {
      case '\n' : /* ignore blanks */
        break;

      case NULL : /* ignore NULLs */
        break;

      case ' ': /* ignore spaces */
        break;

      case EOF: /* End Of File detected */
        break;

      default:
        InputTape[i++] = ch;
        break;
    }
    /****************************
    * init tape to all blanks B
    *****************************/
  }
  printf("Input Tape:\n");
  for(i=0;i < MAX_TAPE_LEN;i++)
    printf("%c",InputTape[i]);
  printf("\n");

  /************
  * close file
  *************/
  if(Tapefp)
    fclose(Tapefp);
  return(0);
}

/*****************************************************
*:: ReadTuringMachine
*
******************************************************/
int  ReadTuringMachine()
{
int Index = 0;
char TMFileName[MAX_FILENAME_LEN];
char FileLine[MAX_FILE_LINE_LEN];
char *ptr;
char *Result;

  /**************************
  * Open Turing Machine file
  ***************************/
  printf("Enter Turing Machine Filename: ");
  gets(TMFileName);
  printf("\n");
  TMfp = fopen(TMFileName,"r");

  /****************************
  * exit if file doesn't exist
  ****************************/
  if(TMfp == NULL)
  {
    printf("No Turing Machine file\n");
    exit(0);
  }

  /************************************************
  * Read and load all of the transition functions.
  *************************************************/
  while(!feof(TMfp))
  {
    Result = fgets(FileLine,MAX_FILE_LINE_LEN,TMfp);
    if(Result == NULL)
       break;
    ptr = FileLine;

    /*********************************************
    * Load transition function into TM structure.
    **********************************************/
    do
    {
       /*****************
       * ignore comments
       ******************/
       if(*ptr == 'C')
         break;
       /*******************************
       * Start of transition function
       ********************************/
       if(*ptr == '(')
       {
         ptr++; /* index to State */
         TM[Index].State =  abs(48 -(int)*ptr);

         ptr++; /* index to comma  */
         ptr++; /* index to symbol */
         TM[Index].Symbol = *ptr;

         /*********************************************
         * Look for right half of transition function.
         *********************************************/
         while(*ptr!='{')
         {
           ptr++;
         }
         ptr++;
         TM[Index].NewState = abs(48 -(int)*ptr); // points to new space
         ptr++; ptr++;
         TM[Index].NewSymbol = *ptr;
         ptr++; ptr++;
         TM[Index].MoveTo = *ptr;
         Index+=1;
       }
    } while (*ptr++ != '\n');
  }
  NumTransitions = Index;
  /*************
  * close file
  *************/
  if(TMfp)
    fclose(TMfp);
  return(0);
}

/******************************************************
*:: ExecuteTransitions:
*
*******************************************************/
int  ExecuteTransitions()
{
static int TapeCtr = 0;
static int TMCtr   = 0;
static char CurrentSymbol = 'B';
static CurrentState = 0;

int Done = 0;
int Found = 0;
int i;

   /***********************************
   * Get start state and staring symbol
   ************************************/
   CurrentState = TM[TMCtr].State;
   CurrentSymbol = TM[TMCtr].Symbol;
   DisplayTransition(CurrentState,TapeCtr);
   printf("\n");

   while(!Done)
   {
     /******************************************
     * Check for match between the transition
     * function symbol and tape symbol.
     *******************************************/
     if(CurrentSymbol == InputTape[TapeCtr])
     {
       /*******************************************
       * Search through transition functions to
       * find a match and write new symbol on tape.
       ********************************************/
       for(TMCtr=0;TMCtr<NumTransitions;TMCtr++)
       {
         if(CurrentState == TM[TMCtr].State)
         {
            CurrentSymbol = TM[TMCtr].Symbol;
            if(CurrentSymbol == InputTape[TapeCtr])
            {
              InputTape[TapeCtr] = TM[TMCtr].NewSymbol;
              CurrentState  = TM[TMCtr].NewState;
              break;
            }
         }
       }
       /******************************************************
       * Move right R or left L according to transition rule
       *******************************************************/
       if(TM[TMCtr].MoveTo == 'R' || TM[TMCtr].MoveTo == 'r')
       {
         if(TapeCtr < MAX_TAPE_LEN)
         {
           TapeCtr+=1;
           if(Mode == SINGLE_STEP)
             PressEnterMsg();
           DisplayTransition(CurrentState,TapeCtr);
         }
         else
         {
           printf("Ran out of tape:RIGHT END -- Turing Machine FAILED\n");
           exit(0);
         }
       }
       else
       {
         if(TM[TMCtr].MoveTo == 'L' || TM[TMCtr].MoveTo == 'l')
         {
           if(TapeCtr > 0)
           {
             TapeCtr-=1;
             if(Mode == SINGLE_STEP)
               PressEnterMsg();
             DisplayTransition(CurrentState,TapeCtr);
           }
           else
           {
             printf("Ran out of Tape:LEFT END -- Turing Machine FAILED\n");
             exit(0);
           }
         } /* if */
       } /* else */
     }  /* if */
     else
     {
       /****************************************
       * search for new transition function at
       *  beginning of list.
       *****************************************/
       Found = 0;
       for(TMCtr=0;TMCtr < NumTransitions;TMCtr++)
       {
         if(CurrentState == TM[TMCtr].State)
         {
            CurrentSymbol = TM[TMCtr].Symbol;
            if(CurrentSymbol == InputTape[TapeCtr])
            {
              Found = 1;
              break;
            }
         } /* if */
         else
         {
           if(TMCtr >= NumTransitions)
           {
             printf("\nTuring Machine HALTED\n");
             Done = 1;
           }
         }
       } /* for */
       if(Found == 0)
       {
         printf("\nTuring Machine HALTED\n");
         Done = 1;
       }
     }
   } /* while */
  return(0);
}

/******************************************************
*:: DisplayTransitions:
*
******************************************************/
int DisplayTransition(int State,int TapePos)
{
int i;

  if(Mode == AUTO)
    printf("\n");

  for(i=0;i<TapePos;i++)
    printf("%c",InputTape[i]);
  printf("(q%d)",State);

  for(i=TapePos;i<MAX_TAPE_LEN;i++)
    printf("%c",InputTape[i]);
  return(0);
}

/*****************************************************
*:: DisplayTape:
*
******************************************************/
int DisplayTape()
{
int i;

  printf("\nFinal tape contents :\n");
  for(i=0;i<MAX_TAPE_LEN;i++)
    printf("%c",InputTape[i]);
  return(0);
}

/*****************************************************
* PressEnterMsg:
******************************************************/
void PressEnterMsg()
{
int InKey;

  printf("\nPress Enter for next transition:");
  InKey = getchar();
}
